/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: ftio.c,v 1.13 2001/02/20 04:43:52 maf Exp $
 */

#if HAVE_CONFIG_H
 #include <config.h>
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <syslog.h>
#include <dirent.h>
#include <limits.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_STRINGS_H
 #include <strings.h>
#endif
#if HAVE_STRING_H
  #include <string.h>
#endif
#include <time.h>
#include <fcntl.h>
#include <zlib.h>
#include "ftlib.h"
#include "fmt.h"

/* XXX TODO
  sequence number checker with ring buffer for flow-capture and flow-receive
  decode split to multiple functions
  document all functions
  encode - jump table
  decode - jump table
*/


/*
 * function: ftio_init
 *
 * Initialize an ftio structure, allocating r/w buffers, zlib, etc.
 * On READ the header is consumed.
 *
 * flags:  FT_IO_FLAG_READ    - setup ftio for reading
 *         FT_IO_FLAG_WRITE   - setup ftio for writing
 *         FT_IO_FLAG_ZINIT   - used with FT_IO_FLAG_WRITE to signal
 *                              future use of compression.
 *         FT_IO_FLAG_NO_SWAP - used with FT_IO_FLAG_WRITE.  Normally
 *                              ftio_write() expects the record in
 *                              host format, and will return it in
 *                              host format.  This will disable the
 *                              swap operation to maximize performance
 *                              in certain cases.
 *
 * ftio_close() must be called on the stream to free resources
 * and flush buffers on WRITE.
 *
 * returns: < 0 error
 *          >= 0 ok
*/
int ftio_init(struct ftio *ftio, int fd, int flag)
{
  int i, ret;

  bzero(ftio, sizeof (struct ftio));

  ftio->fd = fd;

  ret = -1;

  if (flag & FT_IO_FLAG_READ) {

    if (ftheader_read(ftio->fd, &ftio->fth) < 0) {
      fprintf(stderr, "ftheader_read(): failed.\n");
      goto ftio_init_out;
    }

    /* verify stream version */
    if ((ftio->fth.s_version != 1) && (ftio->fth.s_version != 2)) {
      fprintf(stderr, "unsupported stream version\n");
      goto ftio_init_out;
    }
  
    /* fix byte order */
#if BYTE_ORDER == BIG_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_LITTLE_ENDIAN) {
      ftio_header_swap(ftio);
    }
#endif /* BYTE_ORDER == BIG_ENDIAN */
       
#if BYTE_ORDER == LITTLE_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_BIG_ENDIAN) {
      ftio_header_swap(ftio);
    }
#endif /* BYTE_ORDER == LITTLE_ENDIAN */
  
    /* backwards compatability hack */
    if ((ftio->fth.s_version == 1) && (ftio->fth.d_version == 65535))
      ftio->fth.d_version = 1;
  
    /* alloc z_buf if compression set */
    if (ftio->fth.flags & FT_HEADER_FLAG_COMPRESS) {
      if (!(ftio->z_buf = (char*)malloc(FT_Z_BUFSIZE))) {
        fprintf(stderr, "malloc(): failed\n");
        goto ftio_init_out;
      }
    }
  
    /* calculate record size */
    if ((ftio->rec_size = ftio_rec_size(ftio)) < 0) {
      fprintf(stderr, "unsupported record type\n");
      goto ftio_init_out;
    }
  
    /* set byte swap function */
    if (!(ftio->swapf = ftio_rec_swapfunc(ftio))) {
      goto ftio_init_out;
    }
  
    /* alloc d_buf -- 1 for compressed strems, many for uncompressed */
    if (ftio->fth.flags & FT_HEADER_FLAG_COMPRESS)
      i = ftio->rec_size;
    else
      i = FT_D_BUFSIZE;
    if (!(ftio->d_buf = (char*)malloc(i))) {
      fprintf(stderr, "malloc(): failed\n");
      goto ftio_init_out;
    }
  
    /* initialize zlib and set zlib initialized flag */
    if (ftio->fth.flags & FT_HEADER_FLAG_COMPRESS) {
  
      ftio->zs.zalloc = (alloc_func)0;
      ftio->zs.zfree = (free_func)0;
      ftio->zs.opaque = (voidpf)0;
  
      if (inflateInit(&ftio->zs) != Z_OK) {
        fprintf(stderr, "inflateInit(): failed\n");
        goto ftio_init_out;
      }
  
      ftio->flags |= FT_IO_FLAG_ZINIT;
  
      ftio->zs.avail_out = ftio->rec_size;
      ftio->zs.next_out = (Bytef*)ftio->d_buf;

    }
  
    /* mark stream for reading */
    ftio->flags |= FT_IO_FLAG_READ;
  
    ret = 0;

  } else if (flag & FT_IO_FLAG_WRITE) {

#if BYTE_ORDER == LITTLE_ENDIAN
    ftio->fth.byte_order = FT_HEADER_LITTLE_ENDIAN;
#endif /* LITTLE_ENDIAN */

#if BYTE_ORDER == BIG_ENDIAN
    ftio->fth.byte_order = FT_HEADER_BIG_ENDIAN;
#endif /* BIG_ENDIAN */

    /* alloc z_buf if compression set */
    if (flag & FT_IO_FLAG_ZINIT) {

      if (!(ftio->z_buf = (char*)malloc(FT_Z_BUFSIZE))) {
        fprintf(stderr, "malloc(): failed\n");
        goto ftio_init_out;
      }

      ftio->zs.zalloc = (alloc_func)0;
      ftio->zs.zfree = (free_func)0;
      ftio->zs.opaque = (voidpf)0;

      if (deflateInit(&ftio->zs, ftio->z_level) != Z_OK) {
        fprintf(stderr, "deflateInit(): failed\n");
        goto ftio_init_out;
      }

      ftio->flags |= FT_IO_FLAG_ZINIT;
      ftio->fth.flags |= FT_HEADER_FLAG_COMPRESS;

      ftio->zs.next_out = (Bytef*)ftio->z_buf;
      ftio->zs.avail_out = FT_Z_BUFSIZE;

    /* no compression */
    } else {

      if (!(ftio->d_buf = (char*)malloc(FT_D_BUFSIZE))) {
        fprintf(stderr, "malloc(): failed\n");
        goto ftio_init_out;
      }

      ftio->d_end = FT_D_BUFSIZE;

    }

    /* mark stream for writing */
    ftio->flags |= FT_IO_FLAG_WRITE;

    ret = 0;

  } /* write */
  
ftio_init_out:
  
  if (ret) {

    if (ftio->z_buf)
      free (ftio->z_buf);

    if (ftio->d_buf)
      free (ftio->d_buf);

    if (ftio->flags & FT_IO_FLAG_ZINIT)
      inflateEnd(&ftio->zs);

  } /* error */

  return ret;
}

/*
 * function: ftio_set_z_level
 *
 * Set the zlib compression level for a ftio stream.
*/
void ftio_set_z_level(struct ftio *ftio, int z_level)
{

  if ((ftio->fth.flags & FT_HEADER_FLAG_COMPRESS) && (!z_level)) {
    fprintf(stderr, "Compression can not be disabled.\n");
    return;
  }
  
  if ((!(ftio->fth.flags & FT_HEADER_FLAG_COMPRESS)) && (z_level)) {
    fprintf(stderr, "Compression can not be enabled.\n");
    return;
  }

  ftio->z_level = z_level;

  if (z_level) 
    if (deflateParams(&ftio->zs, ftio->z_level, Z_DEFAULT_STRATEGY) != Z_OK)
      fprintf(stderr, "deflateParams(): failed\n");

}
 

/*
 * function: ftio_set_nflows
 *
 * Set the # of flows for a ftio stream header
*/
void ftio_set_nflows(struct ftio *ftio, u_int32 n)
{
  ftio->fth.nflows = n;
}

/*
 * function: ftio_set_streaming
 *
 * Set the streaming flag for a ftio stream
*/
void ftio_set_streaming(struct ftio *ftio, int flag)
{
  if (flag)
    ftio->fth.flags |= FT_HEADER_FLAG_STREAMING;
  else
    ftio->fth.flags &= ~FT_HEADER_FLAG_STREAMING;
}

/*
 * function: ftio_set_ver
 *
 * Set the version information for a ftio stream
*/
int ftio_set_ver(struct ftio *ftio, struct ftver *ver)
{

  ftio->fth.d_version = ver->d_version;
  ftio->fth.s_version = ver->s_version;
  ftio->fth.agg_method = ver->agg_method;
  ftio->fth.agg_version = ver->agg_version;

  /* calculate record size */
  if ((ftio->rec_size = ftio_rec_size(ftio)) < 0) {
    fprintf(stderr, "unsupported record type\n");
    ftio->fth.d_version = 0;
    return -1;
  }

  /* set byte swap function */
  if (!(ftio->swapf = ftio_rec_swapfunc(ftio))) {
    return -1;
  }

  return 0;

}

/*
 * function: ftio_set_byte_order
 *
 * Set the byte order for a ftio stream
*/
void ftio_set_byte_order(struct ftio *ftio, int byte_order)
{
  ftio->fth.byte_order = byte_order;
}

/*
 * function: ftio_set_debug
 *
 * Set the debug level for a ftio stream
*/
void ftio_set_debug(struct ftio *ftio, int debug)
{
  ftio->debug = debug;
}

/*
 * function: ftio_set_comment
 *
 * Set the header comment for a ftio stream
*/
void ftio_set_comment(struct ftio *ftio, char *comment)
{
  strncpy(ftio->fth.comments, comment, FT_HEADER_CMNT_LEN);
  ftio->fth.comments[FT_HEADER_CMNT_LEN-1] = 0;
}

/*
 * function: ftio_set_hostname
 *
 * Set the header hostname for a ftio stream
*/
void ftio_set_hostname(struct ftio *ftio, char *hostname)
{
  strncpy(ftio->fth.hostname, hostname, FT_HEADER_HN_LEN);
  ftio->fth.hostname[FT_HEADER_HN_LEN-1] = 0;
}

/*
 * function: ftio_set_corrupt
 *
 * Set the pdu_corrupt header field
*/
void ftio_set_corrupt(struct ftio *ftio, u_int32 corrupt)
{
  ftio->fth.pdu_corrupt = corrupt;
}

/*
 * function: ftio_set_time
 *
 * Set the header time for a ftio stream
*/
void ftio_set_time(struct ftio *ftio, u_int32 start, u_int32 end)
{
  ftio->fth.start = start;
  ftio->fth.end = end;
}

/*
 * function: ftio_get_ver
 *
 * Get the version from a ftio stream
*/
void ftio_get_ver(struct ftio *ftio, struct ftver *ver)
{
  ver->d_version = ftio->fth.d_version;
  ver->s_version = ftio->fth.s_version;
  ver->agg_method = ftio->fth.agg_method;
  ver->agg_version = ftio->fth.agg_version;
}

/*
 * function: ftio_get_stime
 *
 * Get the starting time from a ftio stream
*/
u_int32 ftio_get_stime(struct ftio *ftio)
{
  return ftio->fth.start;
}

/*
 * function: ftio_get_etime
 *
 * Get the ending time from a ftio stream
*/
u_int32 ftio_get_etime(struct ftio *ftio)
{
  return ftio->fth.end;
}

/*
 * function: ftio_get_rec_total
 *
 * Get the total records processed from a ftio stream
*/
u_int64 ftio_get_rec_total(struct ftio *ftio)
{
  return ftio->rec_total;
}

/*
 * function: ftio_get_debug
 *
 * Get the debug level from a ftio stream
*/
int ftio_get_debug(struct ftio *ftio)
{
  return ftio->debug;
}

/*
 * function: ftio_get_hostname
 *
 * Get the header hostname from a ftio stream
*/
char *ftio_get_hostname(struct ftio *ftio)
{
  return ftio->fth.hostname;
}

/*
 * function: ftio_get_comment
 *
 * Get the header comment from a ftio stream
*/
char *ftio_get_comment(struct ftio *ftio)
{
  return ftio->fth.comments;
}

/*
 * function: ftio_close
 *
 * Free resources allocated with ftio_init()
 * Flush any non empty buffers if stream was initialized for WRITE
 * 
 * returns: <0   error
 *          >= 0 okay
*/
int ftio_close(struct ftio *ftio)
{
  extern int errno;
  int ret, err, n, nbytes;

  ret = -1;
  n = 0;

  if (ftio->flags & FT_IO_FLAG_READ) {

    if (ftio->flags & FT_IO_FLAG_ZINIT)
      inflateEnd(&ftio->zs);

    if (ftio->z_buf)
      free (ftio->z_buf);

    if (ftio->d_buf)
      free (ftio->d_buf);

  } else if (ftio->flags & FT_IO_FLAG_WRITE) {

    /* compression enabled? */
    if (ftio->flags & FT_IO_FLAG_ZINIT) {

      ftio->zs.avail_in = 0;
        
      while (1) {

        err = deflate(&ftio->zs, Z_FINISH);

        /* if done compressing, do final write to disk */
        if (err == Z_STREAM_END)
          break;
    
        /* if anything other than Z_OK, then it's an error */
        if (err != Z_OK) {
          fprintf(stderr, "deflate(): failed\n");
          goto ftio_close_out;
        }   
    
        /* need to flush */
        if (!ftio->zs.avail_out) {

          n = writen(ftio->fd, ftio->z_buf, FT_Z_BUFSIZE);
          nbytes += n;

          if (n < 0) {
            fprintf(stderr, "writen(): %s\n", strerror(errno));
            goto ftio_close_out;
          }
  
          if (n == 0) {
            fprintf(stderr, "writen(): EOF\n");
            goto ftio_close_out;
          }

          ftio->zs.next_out = (Bytef*)ftio->z_buf;
          ftio->zs.avail_out = FT_Z_BUFSIZE;

        } else
          break;
      } /* while 1 */

      n = writen(ftio->fd, ftio->z_buf, FT_Z_BUFSIZE-ftio->zs.avail_out);

      if (n < 0) {
        fprintf(stderr, "writen(): %s\n", strerror(errno));
        goto ftio_close_out;
      }
  
      if (n == 0) {
        fprintf(stderr, "writen(): EOF\n");
        goto ftio_close_out;
      }

      nbytes += n;

      deflateEnd(&ftio->zs);
      ftio->flags &= ~FT_IO_FLAG_ZINIT;

      free(ftio->z_buf);

      ret = nbytes;
    /* no compression */
    } else {

      if (ftio->d_start) {

        n = writen(ftio->fd, ftio->d_buf, ftio->d_start);

        if (n < 0) {
          fprintf(stderr, "writen(): %s\n", strerror(errno));
          goto ftio_close_out;
        }
  
        if (n == 0) {
          fprintf(stderr, "writen(): EOF\n");
          goto ftio_close_out;
        }

        ftio->d_start = 0;

        ret = n;

      } /* buffer not empty */

      ret = 0;

      free (ftio->d_buf);

    } /* compression */

  } /* io stream enabled for write */

  ret = close(ftio->fd);

  /* return bytes written if no error */
  if (ret >= 0)
    ret = n;

ftio_close_out:

  return ret;
  
}

/*
 * function: ftio_zstat_print
 *
 * Print resources utilization associated with ftio zlib usage
 *
 * returns: <0   error
 *          >= 0 okay
*/
void ftio_zstat_print(struct ftio *ftio, FILE *std)
{

  double d;

  d = ((double)ftio->zs.total_out) / ((double)ftio->zs.total_in);
  fprintf(std, "compression: total_in=%lu total_out=%lu  %3.3f:1\n",
    ftio->zs.total_in, ftio->zs.total_out, d);

}

/*
 * function: ftio_read
 *
 * Return the next fts2rec_* in the ftio stream, or 0L for EOF
 *
 * Record is returned in host byte order
 *
 * Stream must be first initialized with ftio_init() 
 *
*/
void *ftio_read(struct ftio *ftio)
{

  extern int errno;
  int n, err;
  void *ret;
  struct fts1rec_compat *compat;

  ret = (void*)0L;

  /* processed compressed stream */
  if (ftio->fth.flags & FT_HEADER_FLAG_COMPRESS) {

    while (1) {

      /*
       * if the inflate buffer is empty, perform a read()
      */

      if (!ftio->zs.avail_in) {

        n = read(ftio->fd, (char*)ftio->z_buf, FT_Z_BUFSIZE);

        /* EOF and inflate buffer is empty -- done. */
        if (!n) {

          /*
           * check for partial record inflated.  This would never
           * happen on a uncorrupted stream
           */
          if (ftio->zs.avail_out != ftio->rec_size)
            fprintf(stderr, "warning, partial inflated record before EOF\n");

          /* signal EOF to caller */
          goto ftio_read_out;

        }

        /* read error -- done. */
        if (n == -1) {
          fprintf(stderr, "read(): %s\n", strerror(errno));
          goto ftio_read_out;
        }

        ftio->zs.avail_in = n;
        ftio->zs.next_in = (Bytef*)ftio->z_buf;

      } /* if inflate buffer empty */

      /*
       * inflate stream, attempt to get a record
       */

      err = inflate(&ftio->zs, Z_SYNC_FLUSH);

      if ((err != Z_OK) && (err != Z_STREAM_END)) {
        fprintf(stderr, "inflate(): failed\n");
        goto ftio_read_out;
      }

      /* if avail_out == 0, then a full record was inflated -- return it */
      if (!ftio->zs.avail_out) {

        /* reset zlib for next call */
        ftio->zs.avail_out = ftio->rec_size;
        ftio->zs.next_out = (Bytef*)ftio->d_buf;

        ret = (void*)ftio->d_buf;
        goto ftio_read_out;

      } /* bytes available for inflate */
    } /* while 1 */
  } /* compressed stream */

  /*
   * uncompressed stream 
   */

  /*
   * if there are not enough bytes between the start pointer and the end
   * of d_buf for a full record, perform a read()
   */

  /* while(1) loop to optimize for normal case of returning bytes from d_buf */
  while (1) {

    n = ftio->d_end - ftio->d_start;

    /* enough bytes in d_buf to return a record? */
    if (n >= ftio->rec_size) {

      n = ftio->d_start;
      ftio->d_start += ftio->rec_size;

      ret = (void*)ftio->d_buf+n;
      goto ftio_read_out;

    /* no, perform a read() to try for more */
    } else {

      /* move trailing partial record to top of buffer */
      if (n)
        bcopy(ftio->d_buf+ftio->d_start, ftio->d_buf, n);

      ftio->d_end = n;
      ftio->d_start = 0;

      n = read(ftio->fd, (char*)ftio->d_buf+ftio->d_end,
        FT_D_BUFSIZE - ftio->d_end);

      /* read failed? */
      if (n < 0) {
        fprintf(stderr, "read(): %s\n", strerror(errno));
        goto ftio_read_out;
      }

      /* eof? */
      if (n == 0) {
        if (ftio->d_start)
          fprintf(stderr, "warning, partial record before EOF\n");
        goto ftio_read_out;
      }

      ftio->d_end += n;

    } /* need a read() */
  } /* while 1 */

ftio_read_out:

  if (ret) {

      /* fix byte ordering */
#if BYTE_ORDER == BIG_ENDIAN
      if (ftio->fth.byte_order == FT_HEADER_LITTLE_ENDIAN)
        ftio->swapf((void*)ret);
#endif /* BYTE_ORDER == BIG_ENDIAN */
  
#if BYTE_ORDER == LITTLE_ENDIAN
      if (ftio->fth.byte_order == FT_HEADER_BIG_ENDIAN)
        ftio->swapf((void*)ret);
#endif /* BYTE_ORDER == LITTLE_ENDIAN */

    /* increment total records processed */
    ftio->rec_total ++;

    /*
     * backwards compatability hack.  Map the stream version 1 into
     * a stream version 2 
     */
    if (ftio->fth.s_version == 1) {

      ftio->fth.flags |= FT_HEADER_FLAG_XLATE;

      if (ftio->fth.d_version == 1) {

        compat = ret;

        ftio->compat_v1.sysUpTime = 0;
        ftio->compat_v1.unix_secs = compat->unix_secs;
        ftio->compat_v1.unix_nsecs = compat->unix_msecs * 1000000;
        ftio->compat_v1.srcaddr = compat->srcaddr;
        ftio->compat_v1.dstaddr = compat->dstaddr;
        ftio->compat_v1.nexthop = compat->nexthop;
        ftio->compat_v1.input = compat->input;
        ftio->compat_v1.output = compat->output;
        ftio->compat_v1.dPkts = compat->dPkts;
        ftio->compat_v1.dOctets = compat->dOctets;
        ftio->compat_v1.Last = compat->Last;
        ftio->compat_v1.First = compat->First;
        ftio->compat_v1.srcport = compat->srcport;
        ftio->compat_v1.dstport = compat->dstport;
        ftio->compat_v1.prot = compat->prot;
        ftio->compat_v1.tos = compat->tos;
        ftio->compat_v1.tcp_flags = compat->flags;

        ret = (void*)&ftio->compat_v1;

      } else if (ftio->fth.d_version == 5) {

        compat = ret;

        ftio->compat_v5.sysUpTime = 0;
        ftio->compat_v5.unix_secs = compat->unix_secs;
        ftio->compat_v5.unix_nsecs = compat->unix_msecs * 1000000;
        ftio->compat_v5.srcaddr = compat->srcaddr;
        ftio->compat_v5.dstaddr = compat->dstaddr;
        ftio->compat_v5.nexthop = compat->nexthop;
        ftio->compat_v5.input = compat->input;
        ftio->compat_v5.output = compat->output;
        ftio->compat_v5.dPkts = compat->dPkts;
        ftio->compat_v5.dOctets = compat->dOctets;
        ftio->compat_v5.Last = compat->Last;
        ftio->compat_v5.First = compat->First;
        ftio->compat_v5.srcport = compat->srcport;
        ftio->compat_v5.dstport = compat->dstport;
        ftio->compat_v5.prot = compat->prot;
        ftio->compat_v5.tos = compat->tos;
        ftio->compat_v5.tcp_flags = compat->flags;
        ftio->compat_v5.src_as = compat->src_as;
        ftio->compat_v5.dst_as = compat->dst_as;
        ftio->compat_v5.src_mask = compat->src_mask;
        ftio->compat_v5.dst_mask = compat->dst_mask;

        ret = (void*)&ftio->compat_v5;

      } /* v5 compat */
    } /* need compat hack */

  } /* ret is set */

  return ret;

}

/*
 * function: ftio_write_header
 *
 * A ftio stream consists of a header and n records.  ftio_write_header()
 * must be called before ftio_write() to output the header portion of
 * the stream
 *
 * Stream must be first initialized with ftio_init() 
 *
 * returns: <0   error
 *          >= 0 okay
 *
*/
int ftio_write_header(struct ftio *ftio)
{

  extern int errno;
  int n, ret;
  int restore;

  ret = -1;
  restore = 0;

  /* if this is not the first time, rewind */
  if (ftio->flags & FT_IO_FLAG_HEADER_DONE) {

    if (lseek(ftio->fd, (off_t)0L, SEEK_SET) == -1) {
      fprintf(stderr, "lseek(): %s\n", strerror(errno));
      goto ftio_write_header_out;
    }

    /* flag to seek back to end of file */
    restore = 1;

    /* mark the file as complete.  Assume that the second call to write_header
     * is actually the last
     */
    ftio->fth.flags |= FT_HEADER_FLAG_DONE;

  }

  ftio->fth.magic1 = FT_HEADER_MAGIC1;
  ftio->fth.magic2 = FT_HEADER_MAGIC2;

  ftio->fth.s_version = 2;

  if ((!ftio->fth.d_version) || (!ftio->fth.byte_order)) {
    fprintf(stderr, "set d_version and byte_order first\n");
    goto ftio_write_header_out;
  }

  if (!(ftio->flags & FT_IO_FLAG_WRITE)) {
    fprintf(stderr, "stream not initialized for writing\n");
    goto ftio_write_header_out;
  }

#if BYTE_ORDER == BIG_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_LITTLE_ENDIAN) {
      ftio_header_swap(ftio);
    }
#endif /* BYTE_ORDER == BIG_ENDIAN */

#if BYTE_ORDER == LITTLE_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_BIG_ENDIAN) {
      ftio_header_swap(ftio);
    }
#endif /* BYTE_ORDER == LITTLE_ENDIAN */

  n = writen(ftio->fd, &ftio->fth, sizeof (struct fts2header));

#if BYTE_ORDER == LITTLE_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_BIG_ENDIAN) {
      ftio_header_swap(ftio);
    }
#endif /* BYTE_ORDER == LITTLE_ENDIAN */

#if BYTE_ORDER == BIG_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_LITTLE_ENDIAN) {
      ftio_header_swap(ftio);
    }
#endif /* BYTE_ORDER == BIG_ENDIAN */

  if (n < 0) {
    fprintf(stderr, "writen(): %s\n", strerror(errno));
    goto ftio_write_header_out;
  }

  if (n == 0) {
    fprintf(stderr, "writen(): EOF\n");
    goto ftio_write_header_out;
  }

  ftio->flags |= FT_IO_FLAG_HEADER_DONE;

ftio_write_header_out:

  if (restore) {

    if (lseek(ftio->fd, (off_t)0L, SEEK_END) == -1) {
      fprintf(stderr, "lseek(): %s\n", strerror(errno));
    }

  }

  ret = n;

  return ret;

}

/*
 * function: ftio_write
 *
 * Schedule fts2rec_* for output.  ftio_write_header() must be called
 * on a stream before ftio_write().  If ftio_close() is not called
 * records may not be written.
 *
 * Stream must be first initialized with ftio_init() 
 *
 * returns: <0   error
 *          >= 0 okay
 *
*/
int ftio_write(struct ftio *ftio, void *data)
{
  extern int errno;
  int ret, n;

  ret = -1;
 
  if (!ftio->flags & FT_IO_FLAG_NO_SWAP) {
#if BYTE_ORDER == BIG_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_LITTLE_ENDIAN)
      ftio->swapf((void*)data);
#endif /* BYTE_ORDER == BIG_ENDIAN */

#if BYTE_ORDER == LITTLE_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_BIG_ENDIAN)
      ftio->swapf((void*)data);
#endif /* BYTE_ORDER == LITTLE_ENDIAN */
  }

  /* compressed stream? */
  if (ftio->fth.flags & FT_HEADER_FLAG_COMPRESS) {

    ftio->zs.next_in = (Bytef*)data;
    ftio->zs.avail_in = ftio->rec_size;

    while (1) {

      if (deflate(&ftio->zs, Z_NO_FLUSH) != Z_OK) {
        fprintf(stderr, "deflate(): failed\n");
        goto ftio_write_out;
      }

      /* need to flush */
      if (!ftio->zs.avail_out) {

        n = writen(ftio->fd, ftio->z_buf, FT_Z_BUFSIZE);

        if (n < 0) {
          fprintf(stderr, "writen(): %s\n", strerror(errno));
          goto ftio_write_out;
        }

        if (n == 0) {
          fprintf(stderr, "writen(): EOF\n");
          goto ftio_write_out;
        }

        ftio->zs.next_out = (Bytef*)ftio->z_buf;
        ftio->zs.avail_out = FT_Z_BUFSIZE;

        ret = n;

      } else {
        ret = ftio->rec_size;
        break;
      }
    } /* deflating */
  /* no, uncompressed stream */
  } else {

    /* flush full buffer */
    if ((ftio->d_start + ftio->rec_size) > ftio->d_end) {

      n = writen(ftio->fd, ftio->d_buf, ftio->d_start);

      if (n < 0) {
        fprintf(stderr, "writen(): %s\n", strerror(errno));
        goto ftio_write_out;
      }

      if (n == 0) {
        fprintf(stderr, "writen(): EOF\n");
        goto ftio_write_out;
      }

      ftio->d_start = 0;

    }

    bcopy(data, ftio->d_buf+ftio->d_start, ftio->rec_size);

    ftio->d_start += ftio->rec_size;

    ret = ftio->rec_size;

  }

ftio_write_out:

  if (!ftio->flags & FT_IO_FLAG_NO_SWAP) {
#if BYTE_ORDER == BIG_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_LITTLE_ENDIAN)
      ftio->swapf((void*)data);
#endif /* BYTE_ORDER == BIG_ENDIAN */

#if BYTE_ORDER == LITTLE_ENDIAN
    if (ftio->fth.byte_order == FT_HEADER_BIG_ENDIAN)
      ftio->swapf((void*)data);
#endif /* BYTE_ORDER == LITTLE_ENDIAN */
  }

  return ret;

}

/*
 * function: ftio_header_print
 *
 * Dump ftio header in readable format
 *
 * Stream must be first initialized with ftio_init() 
 *
*/
void ftio_header_print(struct ftio *ftio, FILE *std, char cc)
{

  struct fts2header *fth;
  char agg_ver, agg_method;
  char *agg_name;

  fth = &ftio->fth;

  if (ftio->fth.flags & FT_HEADER_FLAG_STREAMING)
    fprintf(std, "%c\n%c mode:                 streaming\n", cc, cc);
  else
    fprintf(std, "%c\n%c mode:                 normal\n", cc, cc);

  if (ftio->fth.flags & FT_HEADER_FLAG_XLATE)
    fprintf(std, "%c translated            yes\n", cc);

  if (!(ftio->fth.flags & FT_HEADER_FLAG_STREAMING))
    fprintf(std, "%c capture hostname:     %s\n", cc, fth->hostname);

  if (!(ftio->fth.flags & FT_HEADER_FLAG_STREAMING))
    fprintf(std, "%c capture start:        %s", cc,
      ctime((time_t*)&fth->start));

  fprintf(std, "%c compress:             %s\n", cc, 
    (fth->flags & FT_HEADER_FLAG_COMPRESS) ? "on" : "off");

  fprintf(std, "%c byte order:           ", cc);
  if (fth->byte_order == FT_HEADER_LITTLE_ENDIAN)
    fprintf(std, "little\n");
  else if (fth->byte_order == FT_HEADER_BIG_ENDIAN)
    fprintf(std, "big\n");
  else
    fprintf(std, "BROKEN\n");

  if (!(ftio->fth.flags & FT_HEADER_FLAG_STREAMING))
    fprintf(std, "%c multiple pdu types:   %s\n", cc,
      (fth->flags & FT_HEADER_FLAG_MULT_PDU) ? "yes" : "no");

  fprintf(std, "%c stream version:       %u\n", cc, (int)fth->s_version);
  fprintf(std, "%c export version:       %u\n", cc, (int)fth->d_version);

  if (fth->d_version == 8) {

    agg_ver = ftio->fth.agg_version;
    agg_method = ftio->fth.agg_method;

    if (agg_method == 1)
      agg_name = "AS";
    else if (agg_method == 2)
      agg_name = "Protocol Port";
    else if (agg_method == 3)
      agg_name = "Source Prefix";
    else if (agg_method == 4)
      agg_name = "Destination Prefix";
    else if (agg_method == 5)
      agg_name = "Prefix";
    else
      agg_name = "Unknown";

    fprintf(std, "%c export agg_version:   %u\n", cc, (int)agg_ver);
    fprintf(std, "%c export agg_method:    %u (%s)\n", cc, (int)agg_method,
      agg_name);
  }

  if (!(ftio->fth.flags & FT_HEADER_FLAG_STREAMING))
    fprintf(std, "%c pdu lost      :       %lu\n", cc, (u_long)fth->pdu_lost);

  if (!(ftio->fth.flags & FT_HEADER_FLAG_STREAMING))
    fprintf(std,
      "%c pdu misordered:       %lu\n", cc, (u_long)fth->pdu_misordered);

  if (!(ftio->fth.flags & FT_HEADER_FLAG_STREAMING))
    fprintf(std,
      "%c pdu corrupt:          %lu\n", cc, (u_long)fth->pdu_corrupt);

  if (fth->comments[0])
    fprintf(std, "%c comments:             %s\n", cc, fth->comments);
  
  if (!(ftio->fth.flags & FT_HEADER_FLAG_STREAMING)) {

    if (fth->flags & FT_HEADER_FLAG_DONE) {
      fprintf(std, "%c capture end:          %s", cc,
        ctime((time_t*)&fth->end));
      fprintf(std, "%c capture nflows:       %lu\n", cc,
        (unsigned long)fth->nflows);
    } else
      fprintf(std, "%c note, incomplete flow file\n", cc);
  }

  fprintf(std, "%c\n", cc);

} /* ftio_header_print */

/*
 * function: ftio_rec_swapfunc
 *
 * Return the function required to swap a record.  Used to create
 * jump table based on the d_version and agg_method
 *
*/
void *ftio_rec_swapfunc(struct ftio *ftio)
{

  u_int8 s_ver, d_ver, agg_ver, agg_method;
  void *ret;

  s_ver = ftio->fth.s_version;
  d_ver = ftio->fth.d_version;

  agg_ver = ftio->fth.agg_version;
  agg_method = ftio->fth.agg_method;

  switch (s_ver) {

    case 1:
      ret = fts1rec_swap_compat;
      break;

    case 2:
      switch (ftio->fth.d_version) {

        case 1:
          ret = fts2rec_swap_v1;
          break;

        case 5:
          ret = fts2rec_swap_v5;
          break;

        case 6:
          ret = fts2rec_swap_v6;
          break;

        case 7:
          ret = fts2rec_swap_v7;
          break;

        case 8:
          if (agg_ver != 2) {
            fprintf(stderr, "unsupported agg_version\n");
            ret = (void*)0L;
            break;
          }

          switch (agg_method) {

            case 1:
              ret = fts2rec_swap_v8_1;
              break;

            case 2:
              ret = fts2rec_swap_v8_2;
              break;

            case 3:
              ret = fts2rec_swap_v8_3;
              break;

            case 4:
              ret = fts2rec_swap_v8_4;
              break;

            case 5:
              ret = fts2rec_swap_v8_5;
              break;

            default:
              fprintf(stderr, "unsupported agg_method\n");
              ret = (void*)0L;
              break;

          } /* switch agg_method */
          break;

        default:
          fprintf(stderr, "unsupported d_version\n");
          ret = (void*)0L;
          break;

      } /* switch v8 export */
      break;

    default:
      fprintf(stderr, "unsupported s_version\n");
      ret = (void*)0L;
      break;

  } /* switch s_version */

  return ret;

}

/*
 * function: ftrec_size
 *
 * Return the size of a fts2rec_* based on the d_version and agg_method
 *
*/
int ftrec_size(struct ftver ver)
{

  int ret;

  switch (ver.s_version) {

    case 1:
      ret = sizeof (struct fts1rec_compat);
      break;

    case 2:
      switch (ver.d_version) {

        case 1:
          ret = sizeof (struct fts2rec_v1);
          break;

        case 5:
          ret = sizeof (struct fts2rec_v5);
          break;

        case 6:
          ret = sizeof (struct fts2rec_v6);
          break;

        case 7:
          ret = sizeof (struct fts2rec_v7);
          break;

        case 8:
          if (ver.agg_version != 2) {
            fprintf(stderr, "unsupported agg_version\n");
            ret = -1;
            break;
          }

          switch (ver.agg_method) {

            case 1:
              ret = sizeof (struct fts2rec_v8_1);
              break;

            case 2:
              ret = sizeof (struct fts2rec_v8_2);
              break;

            case 3:
              ret = sizeof (struct fts2rec_v8_3);
              break;

            case 4:
              ret = sizeof (struct fts2rec_v8_4);
              break;

            case 5:
              ret = sizeof (struct fts2rec_v8_5);
              break;

            default:
              fprintf(stderr, "unsupported agg_method\n");
              ret = -1;
              break;

          } /* switch agg_method */
          break;

        default:
          fprintf(stderr, "unsupported d_version\n");
          ret = -1;
          break;

      } /* switch v8 export */
      break;

    default:
      fprintf(stderr, "unsupported s_version\n");
      ret = -1;
      break;

  } /* switch s_version */

  return ret;

}

/*
 * function: ftio_rec_size
 *
 * Return the size of a fts2rec_* based on the initialized ftio
 * stream.
 *
*/
int ftio_rec_size(struct ftio *ftio)
{

  struct ftver ver;
  ver.d_version = ftio->fth.d_version;
  ver.s_version = ftio->fth.s_version;
  ver.agg_method = ftio->fth.agg_method;
  ver.agg_version = ftio->fth.agg_version;

  return ftrec_size(ver);
}


/*
 * function: readn
 *
 * read()'s n bytes from fd
 * returns # of butes read, or -1 for error
 */
int readn(fd, ptr, nbytes)
register int fd;
register char *ptr;
register int nbytes;
{

  int nleft, nread;

  nleft = nbytes;
  while (nleft > 0) {
      nread = read(fd, ptr, nleft);
      if (nread < 0)
        return nread;
      else if (nread == 0)
        break;

      nleft -= nread;
      ptr += nread;
  }
  return (nbytes - nleft);
} /* readn */


/* From Stevens
 *
 * function: writen
 *
 *  write()'s n bytes to fd.
 *  returns # of bytes written, or -1 for error
 */
int writen(register int fd, register void *ptr, register int nbytes)
{

  int nleft, nwritten;

  nleft = nbytes;
  while (nleft > 0) {
    nwritten = write(fd, ptr, nleft);
    if (nwritten <= 0)
      return(nwritten); /* error */

    nleft -= nwritten;
    ptr += nwritten;
  }
  return(nbytes - nleft);
} /* writen */

/*
 * function: ftheader_read
 *
 * load a ftheader, possibly converting from an older version
 *
 * returns: <0   error
 *          >= 0 okay
*/
int ftheader_read(int fd, struct fts2header *h)
{
  extern int errno;
  int n, len, ret;
  struct ftheader_gen gen;
  struct fts1header *h1;
  struct fts2header *h2;
  char buf[400];

  ret = -1;
  h1 = (struct fts1header*)buf;
  h2 = (struct fts2header*)buf;

  /* read the stream header version area */
  if ((n = readn(fd, (char*)&gen, sizeof gen)) < 0) {
    fprintf(stderr, "read(): %s\n", strerror(errno));
    goto ftheader_read_out;
  }

  if (n != sizeof gen) {
    fprintf(stderr, "short read while loading header top.\n");
    goto ftheader_read_out;
  }
        
  /* verify magic */
  if ((gen.magic1 != FT_HEADER_MAGIC1) ||   
      (gen.magic2 != FT_HEADER_MAGIC2)) {
      fprintf(stderr, "bad magic number\n");
      goto ftheader_read_out;
  }

  /* determine how many bytes to read */
  if (gen.s_version == 1)
    len = (sizeof (struct fts1header)) - sizeof gen;
  else if (gen.s_version == 2)
    len = (sizeof (struct fts2header)) - sizeof gen;
  else {
    fprintf(stderr, "Stream format must be 1 or 2.\n");
      goto ftheader_read_out;
  }

  /* insert the generic part to the top of the buffer */
  bcopy(&gen, buf, sizeof gen);

  /* read the rest of the header */
  if ((n = readn(fd, (char*)buf+sizeof gen, len)) < 0) {
    fprintf(stderr, "read(): %s\n", strerror(errno));
    goto ftheader_read_out;
  }

  if (n != len) {
    fprintf(stderr, "short read while loading header.\n");
    goto ftheader_read_out;
  }

  /* if this is a v1 header, need to convert it to a v2 */
  if (gen.s_version == 1) {

    bzero(h, sizeof (struct fts2header));

    h->magic1 = h1->magic1;
    h->magic2 = h1->magic2;
    h->byte_order = h1->byte_order;
    h->s_version = h1->s_version;
    h->d_version = h1->d_version;
    h->start = h1->start;
    h->end = h1->end;
    h->flags = h1->flags;
    h->rotation = h1->rotation;
    h->nflows = h1->nflows;
    h->pdu_lost = h1->pdu_drops;
    h->pdu_misordered = h1->pdu_misordered;

    bcopy(&h1->hostname, &h->hostname, FT_HEADER_HN_LEN);
    bcopy(&h1->comments, &h->comments, FT_HEADER_CMNT_LEN);

  } else if (gen.s_version == 2) {

    bcopy(h2, h, sizeof (struct fts2header));

  }

  ret = 0;

ftheader_read_out:

  return ret;

} /* ftheader_read */

/*
 * function: ftio_check_generic
 *
 * check if this io stream can be used with the ftstrec_gen
 * pseudo record.  fts2rec_gen overlays the common fields of
 * v1, v5, v6, and v7 formats
 *
 * returns: <0   error
 *          >= 0 okay
*/
int ftio_check_generic(struct ftio *ftio)
{
  struct ftver ver;

  ftio_get_ver(ftio, &ver);
 
  if ((ver.d_version != 1) &&
    (ver.d_version != 5) &&
    (ver.d_version != 6) &&
    (ver.d_version != 7)) {
    fprintf(stderr, "export version %d not supported by format.\n",
      ver.d_version);
    return -1;
  }

  return 0;
} /* ftio_check_generic */

/*
 * function: ftltime
 *
 * Flow exports represent time with a combination of uptime of the
 * router, real time, and offsets from the router uptime.  ftltime
 * converts from the PDU to a standard unix seconds/milliseconds
 * representation
 *
 * returns: struct fttime
*/
struct fttime ftltime(u_int32 sys, u_int32 secs, u_int32 nsecs, u_int32 t)
{

  u_int32 sys_s, sys_m;
  struct fttime ftt;

  /* sysUpTime is in milliseconds, convert to seconds/milliseconds */
  sys_s = sys / 1000;
  sys_m = sys % 1000;

  /* unix seconds/nanoseconds to seconds/milliseconds */
  ftt.secs = secs;
  ftt.msecs = nsecs / 1000000L;

  /* subtract sysUpTime from unix seconds */
  ftt.secs -= sys_s;

  /* borrow a second? */
  if (sys_m > ftt.msecs) {
    -- ftt.secs;
    ftt.msecs += 1000;
  }
  ftt.msecs -= sys_m;

  /* add offset which is in milliseconds */
  ftt.secs += t / 1000;
  ftt.msecs += t % 1000;

  /* fix if milliseconds >= 1000 */
  if (ftt.msecs >= 1000) {
    ftt.msecs -= 1000;
    ftt.secs += 1;
  }

  return ftt;

} /* ftltime */

/*
 * function: ftset_init
 *
 * initialize default settings
 *
 * returns: initialized ftset
*/
void ftset_init(struct ftset *ftset, int z_level)
{

  bzero(ftset, sizeof (struct ftset));
  ftset->z_level = z_level;

#if BYTE_ORDER == BIG_ENDIAN
  ftset->byte_order = FT_HEADER_BIG_ENDIAN;
#endif /* BYTE_ORDER == BIG_ENDIAN */
#if BYTE_ORDER == LITTLE_ENDIAN
  ftset->byte_order = FT_HEADER_LITTLE_ENDIAN;
#endif /* BYTE_ORDER == LITTLE_ENDIAN */

} /* ftset_init */
