/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: flow-capture.c,v 1.24 2001/02/21 22:46:14 maf Exp $
 */

#if HAVE_CONFIG_H
 #include <config.h>
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_STRINGS_H
 #include <strings.h>
#endif
#if HAVE_STRING_H
  #include <string.h>
#endif
#include <time.h>
#include <fcntl.h>
#include <syslog.h>
#include "ftlib.h"
#include "report.h"
#include "support.h"
#ifdef LIBWRAP
#include <tcpd.h>
#endif /* LIBWRAP */



/* TODO
 *
 *  sequence number checking
 *
 *  call external program on file rotation
 *
 *  stats file
 *   time   flows/second
 *   1111   1099   
 *
 *  unlink pidfile after SIGQUIT, or any other exit
 *
 *  The TCP client write() needs to have userland buffering
 *
 *  fix the client
 *
 *  option to lock to specific version/subversion and ignore other PDU's
 *
 *  option to specify source IP address
 *
*/


/* XXX dynamic assignment */
#define MAX_SOCKETBUF 200000

#define CAPTURE_PIDFILE    "/var/run/flow-capture.pid"

#define CLIENT_VALID   1   /* entry in use? */
#define CLIENT_HEADER  2   /* sent stream header yet? */

#define CLIENT_MAXCON  32  /* max # of TCP connections to server */

#define SELECT_TIMEOUT 5   /* 5 seconds */

struct client_rec {
  int fd;
  struct sockaddr_in addr;
  int flags;
  u_int32 con_time;
#ifdef LIBWRAP
  struct request_info tcpd;
#endif /* LIBWRAP */
};

struct client {
  struct client_rec recs[CLIENT_MAXCON+1]; /* +1 is used for error recovery */
  struct sockaddr_in addr;
  int fd;
  int len;
  int num;
};

struct rotate {
  u_int32 next;    /* time of next rotation */
  int    cur;     /* current rotation # */
  int    n;       /* number per day */
};

struct file {
  int fd;                   /* file descriptor */
  char name[MAXPATHLEN+1];  /* name */
  char nname[MAXPATHLEN+1]; /* new name */
  u_int32 time;
  struct ftver ftv;
  off_t nbytes;
  u_int32 nflows;
  u_int32 corrupt;
};

int debug;
int sig_pipe_flag, sig_quit_flag, sig_hup_flag;
void sig_pipe();
void sig_quit();
void sig_hup();
char *progname = "flow-capture\0";
int detach;

char *calc_fname(int done, int ver, int agg_method, u_int32 now);
void usage();
int write_pid (int pid, char *file, unsigned short port);
int calc_rotate (int next, u_int32 *trotate, int *cur);

int main(argc, argv)   
int argc;
char **argv;
{
  extern int errno, detach;
  extern char *optarg;
  struct rotate rot;
  fd_set rfd;
  struct file cap_file;
  int i, n, len, invalid, enable_unlink, offset;
  unsigned int v1, v2;
  u_int32 now;
  pid_t pid;
  char work_dir[MAXPATHLEN+1];
  struct ftexporter ftx;
  struct timeval tv;
  struct ftset ftset;
  struct ftfile_ager fta;
  struct client client;
  struct ftio ftio;
  struct ftpdu ftpdu;

  bzero (&rot, sizeof rot);
  bzero (&cap_file, sizeof cap_file);
  bzero (&ftx, sizeof ftx);
  bzero (&tv, sizeof tv);
  bzero (&work_dir, sizeof work_dir);
  bzero (&fta, sizeof fta);
  bzero (&client, sizeof client);
  bzero (&ftpdu, sizeof ftpdu);

  /* defaults + default compression */
  ftset_init(&ftset, Z_DEFAULT_COMPRESSION);

  /* default timeout waiting for an active fd */
  tv.tv_sec = SELECT_TIMEOUT;

  /* listen for PDU's */
  ftx.cli_addr.sin_family = AF_INET;
  ftx.cli_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  ftx.cli_addr.sin_port = htons(FT_PORT);

  /* listen for clients */
  client.addr.sin_family = AF_INET;
  client.addr.sin_addr.s_addr = htonl(INADDR_ANY);
  client.addr.sin_port = htons(FT_PORT);


  /* default detach from parent */
  detach = 1;

  /* default 95 rotations per day, or every 15 minutes */
  rot.n = 95;

  /* no files initially open */
  cap_file.fd = -1;

  while ((i = getopt(argc, argv, "A:b:c:C:d:De:E:I:n:p:V:w:z:")) != -1)
  
    switch (i) {

    case 'A': /* AS substitution */
      ftset.as_sub = atoi(optarg);
      break;
  
    case 'b': /* output byte order */
      if (!strcasecmp(optarg, "little"))
        ftset.byte_order = FT_HEADER_LITTLE_ENDIAN;
      else if (!strcasecmp(optarg, "big"))
        ftset.byte_order = FT_HEADER_BIG_ENDIAN;
      else {
        fprintf(stderr, "expecting \"big\" or \"little\" at -b\n");
        exit (1);
      }
      break;

    case 'c': /* client enable */
      client.num = atoi(optarg);
      if (client.num > CLIENT_MAXCON) {
        fprintf(stderr, "CLIENT_MAXCON (%d) exceeded\n", CLIENT_MAXCON);
        exit (1);
      }
      break;

    case 'C': /* comment field */
      if (strlen(optarg) > (FT_HEADER_CMNT_LEN-1)) {
        fprintf(stderr, "comments limited to %d bytes.\n",
          (FT_HEADER_CMNT_LEN-1));
        exit (1);
      }
      strcpy(ftset.comments, optarg);
      break;
  
    case 'D': /* daemonize */
      detach = 0;
      break;

    case 'e': /* expire */
      fta.max_files = atoi(optarg);
      break;
  
    case 'E': /* expire bytes */
      if ((fta.max_bytes = scan_size(optarg)) == -1) {
        fprintf(stderr, "error scanning size\n");
        exit (1);
      }
      break;
        
    case 'd': /* debug */
      debug = atoi(optarg);
      break;

    case 'I': /* exporter IP */
      ftx.ip = scan_ip(optarg);
      break;

    case 'n': /* # rotations / day */
      rot.n = atoi(optarg);
      /* no more than 1 rotation per 5 minutes */
      if (rot.n > (288-1)) {
        fprintf(stderr, "rotations limited to every 5 minutes\n");
        exit (1);
      }
      break;
      
    case 'p': /* listen port */
      ftx.port = atoi(optarg);
      client.addr.sin_port = htons(atoi(optarg));
      break;

    case 'V': /* PDU version */
      n = sscanf(optarg, "%u.%u", &v1, &v2);
      if (n == 1) {
        ftx.ftv.s_version = FT_IO_SVERSION;
        ftx.ftv.d_version = v1;
        ftx.ftv.set = 1;
      } else if (n == 2) {
        ftx.ftv.s_version = FT_IO_SVERSION;
        ftx.ftv.d_version = v1;
        ftx.ftv.agg_method = v2;
        ftx.ftv.agg_version = 2;
        ftx.ftv.set = 1;
      } else {
        fprintf(stderr, "Error scanning version.\n");
        exit (1);
      }
      break;
        
    case 'w': /* working directory */
      if (strlen(optarg) > (MAXPATHLEN)) {
        fprintf(stderr, "Pathname too long\n");
        exit (1);
      }
      strcpy(work_dir, optarg);
      break;

    case 'z': /* compress level */
      ftset.z_level = atoi(optarg);
      if ((ftset.z_level < 0) || (ftset.z_level > 9)) {
        fprintf(stderr, "Compression level must be between 0 and 9\n");
        exit (1);
      }
      break;

    default:
      usage();
      exit (1);
      break;
        
    } /* switch */

  /* IP address of exporter is required */
  if (!ftx.ip) {
    fprintf(stderr, "Must specify exporter IP with -I.\n");
    exit (1);
  }

  /* listen port */
  ftx.cli_addr.sin_port = htons(ftx.port);

  /* if debugging is enabled do not unlink any files when aging */
  if (debug) 
    enable_unlink = 0;
  else
    enable_unlink = 1;

 /* daemonize */
  if (detach) {
    if ((pid = fork()) == -1) {
      fprintf(stderr, "fork(): %s\n", strerror(errno));
      exit (1);
    } else if (pid) {
      write_pid(pid, CAPTURE_PIDFILE, ftx.port);
      exit (0); /* parent */
    }

    umask(0022);
    setsid();
    for (n = 0; n < 16; ++n) /* XXX dynamically get NOFILE */
      close (n);

    report_init(0);

  } else {
    report_init(1);
  }

  /* if client attachments are enabled, setup a listener */
  if (client.num) {
    
    /* socket for listen */
    if ((client.fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      report(LOG_ERR, "socket(): %s", strerror(errno));
      exit (1);
    }
      
    if (bind(client.fd, (struct sockaddr*)&client.addr, sizeof (client.addr))
      < 0) {
      report(LOG_ERR, "bind(): %s", strerror(errno));
      exit (1);
    }
      
    /* listen for new TCP connections */
    if (listen(client.fd, 5) < 0) {
      report(LOG_ERR, "listen(): %s", strerror(errno));
      exit (1);
    }

    /* non blocking */
    if (fcntl(client.fd, F_SETFL, O_NONBLOCK) < 0) {
      report(LOG_ERR, "fcntl(): %s", strerror(errno));
      exit (1);
    }
  } /* clients enabled */

  /*
   * configure signal handlers
   */

  if (signal(SIGPIPE, sig_pipe) < 0) {
    report(LOG_ERR, "signal(): %s", strerror(errno));
    exit (1);
  }

  if (signal(SIGHUP, sig_hup) < 0) {
    report(LOG_ERR, "signal(): %s", strerror(errno));
    exit (1);
  }

  if (signal(SIGQUIT, sig_quit) < 0) {
    report(LOG_ERR, "signal(): %s", strerror(errno));
    exit (1);
  }

  /* rotations needs to divide the day evenly */
  if (rot.n && (86400 % (rot.n+1))) {
    report(LOG_ERR, "rotations must divide the day evenly.");
    exit (1);
  }

  /* sandbox */
  if (chdir(work_dir) == -1) {
    report(LOG_ERR, "chdir(%s): %s", work_dir, strerror(errno));
    exit (1);
  }

  /*
   * load directory entries into the file ager
   */
  if (ftfile_scandir(&fta, ".")) {
    report(LOG_ERR, "ftfile_scandir(): failed");
    exit (1);
  }

  /* debugging gets a dump of the ager */
  if (debug)
    ftfile_dump(&fta);

  /* run the ager once now */
  if (ftfile_expire(&fta, enable_unlink, (u_int32)0)) {
    report(LOG_ERR, "ftfile_expire(): failed");
    exit (1);
  }

  /* get hostname */
  if (gethostname((char*)&ftset.hostname, FT_HEADER_HN_LEN-1) == -1) {
    report(LOG_ERR, "gethostname(): %s", strerror(errno));
    exit (1);
  }

  /* ensure null terminated */
  ftset.hostname[FT_HEADER_HN_LEN-1] = 0;

  /* socket to receive flow pdu exports */
  if ((ftx.fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    report(LOG_ERR, "socket(): %s", strerror(errno));
    exit(1);
  }

/* XXX set MAX_SOCKETBUF dynamically */
  /* socket buffer */
  n = MAX_SOCKETBUF;
  if (setsockopt(ftx.fd, SOL_SOCKET, SO_RCVBUF, (char*)&n, sizeof (n)) < 0) {
    report(LOG_ERR, "setsockopt(SO_RCVBUF=%d): %s", n, strerror(errno));
    exit(1);
  }

  if (bind(ftx.fd, (struct sockaddr*)&ftx.cli_addr, sizeof(ftx.cli_addr)) < 0) {
    report(LOG_ERR, "bind(): %s", strerror(errno));
    exit (1);
  }

  while (1) {

    FD_ZERO (&rfd);
    FD_SET (ftx.fd, &rfd);

    if (client.num)
      FD_SET (client.fd, &rfd);

/* XXX shouldn't this be the max of the clients and ftx? */
/* XXX shouldn't and all the client fd's too? */

    if (select (ftx.fd+1, &rfd, (fd_set *)0, (fd_set *)0, &tv) < 0)  {
      if (errno == EINTR) {
        FD_ZERO (&rfd);
      } else {
        report(LOG_ERR, "select(): %s", strerror(errno));
        exit (1);
      }
    }

    bzero (&tv, sizeof tv);
    tv.tv_sec = SELECT_TIMEOUT;

    now = (u_int32)time((time_t)0L); 

    /* new client connection ? */ 
    if (client.num && (FD_ISSET(client.fd, &rfd))) {

      i = -1;
      invalid = 0;

      /* find an open client slot */
      for (n = 0; n < client.num; ++n) {
        if (!(client.recs[n].flags & CLIENT_VALID)) {
          i = n;
          break;
        }
      }

      /* if > max connections, accept it and exit */
      if (i == -1) {
        report(LOG_WARNING, "too many clients");
        i = CLIENT_MAXCON;
        invalid = 1;
      }

      /* accept the connection */
      len = sizeof(client.recs[i].addr);
      client.recs[i].fd =
        accept(client.fd, (struct sockaddr*)&client.recs[i].addr, &len);

      if (client.recs[i].fd < 0)
        report(LOG_ERR, "accept(): %s", strerror(errno));
      else {
        report(LOG_INFO, "new client i=%d IP=%s", i,
          inet_ntoa(client.recs[i].addr.sin_addr));
        client.recs[i].flags = CLIENT_VALID;
        client.recs[i].con_time = now;
      }

#ifdef LIBWRAP
      request_init(&client.tcpd, RQ_DAEMON, "flow-capture", RQ_FILE,
        client.recs[i].fd, NULL);

      fromhost(&client.tcpd);

      if (!hosts_access(&client.tcpd)) {
        report(LOG_WARNING, "refused by libwrap");
        close(client.recs[i].fd);
        client.recs[i].flags = 0;
        invalid = 0;
      }
#endif /* LIBWRAP */

      if (invalid) {
        close(client.recs[i].fd);
        client.recs[i].flags = 0;
      }

    } /* new client */

    /* flag for work later on */
    ftx.ftd.count = 0;
   
    /* PDU ready */
    if (FD_ISSET(ftx.fd, &rfd)) {

      len = sizeof(ftx.srv_addr);

      if ((ftpdu.len = recvfrom(ftx.fd, (char*)&ftpdu.buf,
        sizeof ftpdu.buf, 0, (struct sockaddr*)&ftx.srv_addr,
        &len)) == -1) {
        report(LOG_ERR, "recvfrom(): %s", strerror(errno));
        exit (1);
      }

      if (ftx.ip && ftx.ip != ntohl(ftx.srv_addr.sin_addr.s_addr)) {
        report(LOG_ERR, "src IP (%s) not configured.  Skipping PDU.",
          inet_ntoa(ftx.srv_addr.sin_addr));
        cap_file.corrupt ++;
        goto skip_pdu_decode;
      }


      /* verify integrity, get version */
      if (ftpdu_verify(&ftpdu) < 0) {
        report(LOG_ERR, "ftpdu_verify(): failed.  Skipping PDU.");
        cap_file.corrupt ++;
        goto skip_pdu_decode;
      }

      if (ftpdu_check_seq(&ftpdu) < 0) {
        report(LOG_ERR,
          "ftpdu_seq_check(): failed.  expecting=%ld received=%ld\n",
          (u_long)ftpdu.seq_exp, (u_long)ftpdu.seq_rcv);
      }

      /* first time? */
      if (!ftx.ftv.set) {
  
        /* copy to compare next time */
        bcopy(&ftpdu.ftv, &ftx.ftv, sizeof ftx.ftv);

        /* flag struct as configured */
        ftx.ftv.set = 1;
  
      } else {

        /* ensure previous version == current decoded pdu version */
        if ((ftx.ftv.d_version != ftpdu.ftv.d_version) ||
            (ftx.ftv.agg_method != ftpdu.ftv.agg_method)) {

          report(LOG_ERR, "Unexpected PDU: od=%d d=%d oagg=%d agg=%d\n",
            (int)ftx.ftv.d_version, (int)ftpdu.ftv.d_version,
            (int)ftx.ftv.agg_method, (int)ftpdu.ftv.agg_method);

          cap_file.corrupt ++;

          goto skip_pdu_decode;

        } /* PDU version ok */

      } /* compare versions */

      /* decode the pdu */
      ftx.ftd.byte_order = ftset.byte_order;
      ftx.ftd.as_sub = ftset.as_sub;
      n = fts2rec_pdu_decode(&ftpdu, &ftx.ftd);

    } /* PDU on receive buffer */

skip_pdu_decode:

    /* no current file and pdu version has been set -> create file */
    if ((cap_file.fd == -1) && (ftx.ftv.d_version)) {

      /* calculate the current rotation and next rotate time */
      if (calc_rotate(rot.n, &rot.next, &rot.cur) == -1) {
        report(LOG_ERR, "calc_rotate() failed");
        exit (1);
      }

      /* remember when file was created */
      cap_file.time = now;

      /* remember the version encoded in the filename */
      bcopy(&ftx.ftv, &cap_file.ftv, sizeof cap_file.ftv);

      /* construct the capture file name */
      strcpy(cap_file.name, calc_fname(0, (int)cap_file.ftv.d_version,
        (int)cap_file.ftv.agg_method, cap_file.time));

      /* create/open the capture file */
      if ((cap_file.fd = open(cap_file.name, O_WRONLY|O_CREAT|O_TRUNC, 0644))
        == -1) {
        report(LOG_ERR, "open(%s): %s", cap_file.name, strerror(errno));
        exit (1);
      }

      /* initialize the IO stream */
      if (ftio_init(&ftio, cap_file.fd, FT_IO_FLAG_NO_SWAP | FT_IO_FLAG_WRITE |
        ((ftset.z_level) ? FT_IO_FLAG_ZINIT : 0) ) < 0) {
        report(LOG_ERR, "ftio_init(): failed\n");
        exit (1);
      }

      /* set the version information in the io stream */
      if (ftio_set_ver(&ftio, &ftx.ftv) < 0) {
        report(LOG_ERR, "ftio_set_ver(): failed\n");
        exit(1);
      }

      ftio_set_comment(&ftio, ftset.comments);

      ftio_set_hostname(&ftio, ftset.hostname);

      ftio_set_byte_order(&ftio, ftset.byte_order);

      ftio_set_z_level(&ftio, ftset.z_level);

      ftio_set_time(&ftio, cap_file.time, 0);

      ftio_set_debug(&ftio, debug);

      /* header first */
      if (ftio_write_header(&ftio) < 0) {
        report(LOG_ERR, "ftio_write_header(): failed\n");
        exit (1);
      }
    } /* create capture file and init new io stream */

    /* if the decode buffer has entries write them out */

    for (i = 0, offset = 0; i < ftx.ftd.count;
      ++i, offset += ftx.ftd.rec_size) {
        
      if ((n = ftio_write(&ftio, ftx.ftd.buf+offset)) < 0) {
        report(LOG_ERR, "ftio_write(): failed\n");
        exit (1);
      }

      /* update # of bytes stored in capture file */
      cap_file.nbytes += n;

    }

    /* update # of flows stored in capture file */
    cap_file.nflows += ftx.ftd.count;

    /*
     * time for a new file ?
     */
    if ((now > rot.next) || sig_quit_flag || sig_hup_flag) {
            
      if (cap_file.fd != -1) {
          
        sig_hup_flag = 0; /* re-arm */

        ftio_set_time(&ftio, cap_file.time, now);

        ftio_set_corrupt(&ftio, cap_file.corrupt);

        ftio_set_nflows(&ftio, cap_file.nflows);

        /* re-write header first */
        if (ftio_write_header(&ftio) < 0) {
          report(LOG_ERR, "ftio_write_header(): failed\n");
          exit (1);
        }

        if ((n = ftio_close(&ftio)) < 0) {
          report(LOG_ERR, "ftio_close(): failed\n");
          exit (1);
        }

        cap_file.nbytes += n;

        /* construct final version of capture filename */
        strcpy(cap_file.nname, calc_fname(1, (int)cap_file.ftv.d_version,
          (int)cap_file.ftv.agg_method, cap_file.time));
        
        /* rename working to final */
        if (rename(cap_file.name, cap_file.nname) == -1) {
          report(LOG_ERR, "rename(%s,%s): %s", cap_file.name, cap_file.nname,
            strerror(errno));
          exit (1);
        }

        /* add it to the ager */
        if (ftfile_add_tail(&fta, cap_file.nname, cap_file.nbytes,
          cap_file.time)) {
          report(LOG_ERR, "ftfile_add_tail(%s): failed\n", cap_file.name);
        }

        /* reset */
        bzero(&cap_file, sizeof cap_file);

        /* invalidate file descriptor */
        cap_file.fd = -1;

        /* had enough */
        if (sig_quit_flag)
          goto main_exit;
      } /* file open */
    } /* time for new file */

    if (!(cap_file.nflows % 1000)) {
      if (ftfile_expire(&fta, enable_unlink, cap_file.nbytes)) {
        report(LOG_ERR, "ftfile_expire(): failed");
        exit (1);
      }
    } /* ager run? */

  } /* while 1 */

main_exit:

  return 0;

} /* main */

void usage() {

  fprintf(stderr, "flow capture:\n\n");
  fprintf(stderr, " -A #  replace exports with AS 0 with #\n");
  fprintf(stderr, " -b s  byte order (little or big)\n");
  fprintf(stderr, " -C s  comment field\n");
  fprintf(stderr, " -c #  enable n flow clients\n");
  fprintf(stderr, " -d #  debug level.\n");
  fprintf(stderr, " -D    do not daemonize.\n");
  fprintf(stderr, " -e #  files to keep before expire. (0=disable)\n");
  fprintf(stderr, " -E #  bytes to keep before expire. (0=disable)\n");
  fprintf(stderr, " -I ip IP address of device to capture from (required)\n");
  fprintf(stderr, " -n #  file rotations per day.\n");
  fprintf(stderr, " -p #  UDP port to accept flows on.\n");
  fprintf(stderr, " -s #  size of files to keep before expire. (0=disable)\n"
);
  fprintf(stderr, "       use b,K,M,G as multipliers, ie 16M for 16*1024*1024\n");
  fprintf(stderr, " -V #  version of export to capture.  Use #.# for V8\n");
  fprintf(stderr, "   1   - no sequence #, engine id, or AS\n");
  fprintf(stderr, "   5   - 1+ sequence #, engine id, and AS\n");
  fprintf(stderr, "   6   - 5+ encapsulation size\n");
  fprintf(stderr, "   7   - 5+ router shortcut (Catalyst switches)\n");
  fprintf(stderr, "   8.1 - AS aggregation\n");
  fprintf(stderr, "   8.2 - Protocol Port aggregation\n");
  fprintf(stderr, "   8.3 - Source Prefix aggregation\n");
  fprintf(stderr, "   8.4 - Destination Prefix aggregation\n");
  fprintf(stderr, "   8.5 - Prefix aggregation\n");
  fprintf(stderr, " -w x  working directory pathname.\n");
  fprintf(stderr, " signals\n");
  fprintf(stderr, "   SIGHUP  - close and rotate current file\n");
  fprintf(stderr, "   SIGQUIT - close current file and exit\n");
  fprintf(stderr, "\n\n");
    
} /* usage */ 

int write_pid (int pid, char *file, unsigned short port)
{
  int fd, len;
  char str[16], *c;
  
  if (!(c = (char*)malloc(strlen(file)+16)))
    return -1;

  sprintf(c, "%s.%d", file, (int)port);
    
  len = sprintf(str, "%u\n", (unsigned)pid);
  if ((fd = open(c, O_WRONLY|O_CREAT|O_TRUNC, 0644)) == -1){
    free (c);
    return -1; 
  }
  free (c);
  if (write(fd, str, len) != len) {
    close (fd);
    return -1;
  }
  return close (fd);

} /* write_pid */

void sig_pipe()
{
  sig_pipe_flag = 1;
}
    
void sig_hup()
{
  sig_hup_flag = 1;
}
    
void sig_quit()
{  
  sig_quit_flag = 1;
}

int calc_rotate (int next, u_int32 *trotate, int *cur)
{

  time_t now;       /* current time */
  u_long irotate;   /* interval of next in seconds */
  time_t morning;   /* start of the day */
  struct tm *tm1;
        
  irotate = 86400 / (next+1);
       
  *cur = 0;
  
  if ((now = time((time_t)0L)) == -1)
    return -1;
    
  /* calculate start of the day */
  if (!(tm1 = localtime (&now)))
    return -1;

  tm1->tm_sec = 0;
  tm1->tm_min = 0;
  tm1->tm_hour = 0;
 
  if ((morning = mktime(tm1)) == -1)
    return -1;
  
  if (next)
    *trotate = morning + irotate;
  else  
    *trotate = morning + 86400;
  
  while (now > *trotate) {
  
    ++ *cur;
    *trotate += irotate;
  
  }

  return 0;

} /* calc_rotate */

char *calc_fname(int done, int ver, int agg_method, u_int32 now)
{

  struct tm *tm;
  static char fname[FT_FILELEN+1];
  
  if (!(tm = localtime ((time_t*)&now)))
    return "";

  if (ver == 8) {

    /* ft-vNNmNN.YYYY-DD-MM.HHMMSS */
    if (done)
      sprintf(fname, "ft-v%2.2dm%2.2d.%4.4d-%2.2d-%2.2d.%2.2d%2.2d%2.2d", ver,
        agg_method, (int)tm->tm_year+1900, (int)tm->tm_mon+1, (int)tm->tm_mday,
        (int)tm->tm_hour, (int)tm->tm_min, (int)tm->tm_sec);
    else
      sprintf(fname, "tmp-v%2.2dm%2.2d.%4.4d-%2.2d-%2.2d.%2.2d%2.2d%2.2d", ver,
        agg_method, (int)tm->tm_year+1900, (int)tm->tm_mon+1, (int)tm->tm_mday,
        (int)tm->tm_hour, (int)tm->tm_min, (int)tm->tm_sec);

  } else {
  
    /* ft-vNN.YYYY-DD-MM.HHMMSS */
    if (done)
      sprintf(fname, "ft-v%2.2d.%4.4d-%2.2d-%2.2d.%2.2d%2.2d%2.2d", ver,   
        (int)tm->tm_year+1900, (int)tm->tm_mon+1, (int)tm->tm_mday,
        (int)tm->tm_hour, (int)tm->tm_min, (int)tm->tm_sec);
    else
      sprintf(fname, "tmp-v%2.2d.%4.4d-%2.2d-%2.2d.%2.2d%2.2d%2.2d", ver,
        (int)tm->tm_year+1900, (int)tm->tm_mon+1, (int)tm->tm_mday,
        (int)tm->tm_hour, (int)tm->tm_min, (int)tm->tm_sec);

  } /* ver != 8 */

  return fname;

} /* calc_fname */
